@@ -5,12 +5,19 @@ The ...Adapter classes are responsible for finding adapters.
 The ...Match classes trim the reads.
 """
 import logging
-from enum import Enum
+from enum import IntFlag
 from collections import defaultdict
 from typing import Optional, Tuple, Sequence, Dict, Any, List, Union
 from abc import ABC, abstractmethod
 
-from . import align
+from .align import (
+    EndSkip,
+    Aligner,
+    PrefixComparer,
+    SuffixComparer,
+    edit_environment,
+    hamming_environment,
+)
 
 logger = logging.getLogger()
 
@@ -19,19 +26,22 @@ class InvalidCharacter(Exception):
     pass
 
 
-class Where(Enum):
-    # Constants for the Aligner.locate() function.
-    # The function is called with SEQ1 as the adapter, SEQ2 as the read.
-    # TODO get rid of those constants, use strings instead
-    BACK = align.START_WITHIN_SEQ2 | align.STOP_WITHIN_SEQ2 | align.STOP_WITHIN_SEQ1
-    FRONT = align.START_WITHIN_SEQ2 | align.STOP_WITHIN_SEQ2 | align.START_WITHIN_SEQ1
-    PREFIX = align.STOP_WITHIN_SEQ2
-    SUFFIX = align.START_WITHIN_SEQ2
+# TODO remove this enum, this should be within each Adapter class
+class Where(IntFlag):
+    """
+    Aligner flag combinations for all adapter types.
+
+    "REFERENCE" is the adapter sequence, "QUERY" is the read sequence
+    """
+
+    BACK = EndSkip.QUERY_START | EndSkip.QUERY_STOP | EndSkip.REFERENCE_END
+    FRONT = EndSkip.QUERY_START | EndSkip.QUERY_STOP | EndSkip.REFERENCE_START
+    PREFIX = EndSkip.QUERY_STOP
+    SUFFIX = EndSkip.QUERY_START
     # Just like FRONT/BACK, but without internal matches
-    FRONT_NOT_INTERNAL = align.START_WITHIN_SEQ1 | align.STOP_WITHIN_SEQ2
-    BACK_NOT_INTERNAL = align.START_WITHIN_SEQ2 | align.STOP_WITHIN_SEQ1
-    ANYWHERE = align.SEMIGLOBAL
-    LINKED = 'linked'
+    FRONT_NOT_INTERNAL = EndSkip.REFERENCE_START | EndSkip.QUERY_STOP
+    BACK_NOT_INTERNAL = EndSkip.QUERY_START | EndSkip.REFERENCE_END
+    ANYWHERE = EndSkip.SEMIGLOBAL
 
 
 def returns_defaultdict_int():
@@ -54,7 +64,7 @@ class EndStatistics:
         self.allows_partial_matches: bool = adapter.allows_partial_matches
         # self.errors[l][e] == n iff a sequence of length l matching at e errors was removed n times
         self.errors: Dict[int, Dict[int, int]] = defaultdict(returns_defaultdict_int)
-        self.adjacent_bases = {'A': 0, 'C': 0, 'G': 0, 'T': 0, '': 0}
+        self.adjacent_bases = {"A": 0, "C": 0, "G": 0, "T": 0, "": 0}
         # TODO avoid hard-coding the list of classes
         self._remove_prefix = isinstance(adapter, FrontAdapter)
 
@@ -75,8 +85,8 @@ class EndStatistics:
             or self.effective_length != other.effective_length
             or self.indels != other.indels
         ):
-            raise RuntimeError('Incompatible EndStatistics, cannot be added')
-        for base in ('A', 'C', 'G', 'T', ''):
+            raise RuntimeError("Incompatible EndStatistics, cannot be added")
+        for base in ("A", "C", "G", "T", ""):
             self.adjacent_bases[base] += other.adjacent_bases[base]
         for length, error_dict in other.errors.items():
             for errors in error_dict:
@@ -102,14 +112,14 @@ class EndStatistics:
         # FIXME this is broken for 'anywhere' adapters
         if self._remove_prefix:
             seq = seq[::-1]
-        allowed_bases = 'CGRYSKMBDHVN' if self.has_wildcards else 'GC'
-        p = 1.
+        allowed_bases = "CGRYSKMBDHVN" if self.has_wildcards else "GC"
+        p = 1.0
         probabilities = [p]
         for i, c in enumerate(seq):
             if c in allowed_bases:
-                p *= gc_content / 2.
+                p *= gc_content / 2.0
             else:
-                p *= (1. - gc_content) / 2.
+                p *= (1.0 - gc_content) / 2.0
             probabilities.append(p)
         return probabilities
 
@@ -156,7 +166,6 @@ class SingleAdapterStatistics(AdapterStatistics, ABC):
 
 
 class FrontAdapterStatistics(SingleAdapterStatistics):
-
     def add_match(self, match: "RemoveBeforeMatch"):
         self.end.errors[match.removed_sequence_length()][match.errors] += 1
 
@@ -165,7 +174,6 @@ class FrontAdapterStatistics(SingleAdapterStatistics):
 
 
 class BackAdapterStatistics(SingleAdapterStatistics):
-
     def add_match(self, match: "RemoveAfterMatch"):
         adjacent_base = match.adjacent_base()
         self.end.errors[match.removed_sequence_length()][match.errors] += 1
@@ -209,10 +217,14 @@ class LinkedAdapterStatistics(AdapterStatistics):
     def add_match(self, match: "LinkedMatch"):
         # TODO this is duplicated code
         if match.front_match:
-            self.front.errors[match.front_match.removed_sequence_length()][match.errors] += 1
+            self.front.errors[match.front_match.removed_sequence_length()][
+                match.front_match.errors
+            ] += 1
         if match.back_match:
             adjacent_base = match.back_match.adjacent_base()
-            self.back.errors[match.back_match.removed_sequence_length()][match.errors] += 1
+            self.back.errors[match.back_match.removed_sequence_length()][
+                match.back_match.errors
+            ] += 1
             try:
                 self.back.adjacent_bases[adjacent_base] += 1
             except KeyError:
@@ -281,13 +293,28 @@ class Match(ABC):
     def trimmed(self, read):
         pass
 
+    @abstractmethod
+    def match_sequence(self):
+        pass
+
 
 class SingleMatch(Match, ABC):
     """
     Representation of a single adapter matched to a single string
     """
-    __slots__ = ['astart', 'astop', 'rstart', 'rstop', 'matches', 'errors',
-        'adapter', 'sequence', 'length', 'adjacent_base']
+
+    __slots__ = [
+        "astart",
+        "astop",
+        "rstart",
+        "rstop",
+        "score",
+        "errors",
+        "adapter",
+        "sequence",
+        "length",
+        "adjacent_base",
+    ]
 
     def __init__(
         self,
@@ -295,7 +322,7 @@ class SingleMatch(Match, ABC):
         astop: int,
         rstart: int,
         rstop: int,
-        matches: int,
+        score: int,
         errors: int,
         adapter: "SingleAdapter",
         sequence: str,
@@ -304,7 +331,7 @@ class SingleMatch(Match, ABC):
         self.astop: int = astop
         self.rstart: int = rstart
         self.rstop: int = rstop
-        self.matches: int = matches
+        self.score: int = score
         self.errors: int = errors
         self.adapter: SingleAdapter = adapter
         self.sequence = sequence
@@ -314,8 +341,11 @@ class SingleMatch(Match, ABC):
         self.length: int = astop - astart
 
     def __repr__(self):
-        return 'SingleMatch(astart={}, astop={}, rstart={}, rstop={}, matches={}, errors={})'.format(
-            self.astart, self.astop, self.rstart, self.rstop, self.matches, self.errors)
+        return (
+            f"{self.__class__.__name__}(astart={self.astart}, astop={self.astop}, "
+            f"rstart={self.rstart}, rstop={self.rstop}, "
+            f"score={self.score}, errors={self.errors})"
+        )
 
     def __eq__(self, other) -> bool:
         return (
@@ -324,7 +354,7 @@ class SingleMatch(Match, ABC):
             and self.astop == other.astop
             and self.rstart == other.rstart
             and self.rstop == other.rstop
-            and self.matches == other.matches
+            and self.score == other.score
             and self.errors == other.errors
             and self.adapter is other.adapter
             and self.sequence == other.sequence
@@ -339,10 +369,13 @@ class SingleMatch(Match, ABC):
         If there are indels, this is not reliable as the full alignment
         is not available.
         """
-        wildcards = [self.sequence[self.rstart + i] for i in range(self.length)
-            if self.adapter.sequence[self.astart + i] == wildcard_char and
-                self.rstart + i < len(self.sequence)]
-        return ''.join(wildcards)
+        wildcards = [
+            self.sequence[self.rstart + i]
+            for i in range(self.length)
+            if self.adapter.sequence[self.astart + i] == wildcard_char
+            and self.rstart + i < len(self.sequence)
+        ]
+        return "".join(wildcards)
 
     def get_info_records(self, read) -> List[List]:
         seq = read.sequence
@@ -352,22 +385,25 @@ class SingleMatch(Match, ABC):
             self.errors,
             self.rstart,
             self.rstop,
-            seq[0:self.rstart],
-            seq[self.rstart:self.rstop],
-            seq[self.rstop:],
+            seq[0 : self.rstart],
+            seq[self.rstart : self.rstop],
+            seq[self.rstop :],
             self.adapter.name,
         ]
         if qualities:
             info += [
-                qualities[0:self.rstart],
-                qualities[self.rstart:self.rstop],
-                qualities[self.rstop:],
+                qualities[0 : self.rstart],
+                qualities[self.rstart : self.rstop],
+                qualities[self.rstop :],
             ]
         else:
             info += ["", "", ""]
 
         return [info]
 
+    def match_sequence(self):
+        return self.sequence[self.rstart : self.rstop]
+
     @abstractmethod
     def removed_sequence_length(self) -> int:
         pass
@@ -376,10 +412,6 @@ class SingleMatch(Match, ABC):
 class RemoveBeforeMatch(SingleMatch):
     """A match that removes sequence before the match"""
 
-    def __repr__(self):
-        return 'RemoveBeforeMatch(astart={}, astop={}, rstart={}, rstop={}, matches={}, errors={})'.format(
-            self.astart, self.astop, self.rstart, self.rstop, self.matches, self.errors)
-
     def rest(self) -> str:
         """
         Return the part of the read before this match if this is a
@@ -387,7 +419,7 @@ class RemoveBeforeMatch(SingleMatch):
         return the part after the match if this is not a 'front' adapter (3').
         This can be an empty string.
         """
-        return self.sequence[:self.rstart]
+        return self.sequence[: self.rstart]
 
     def remainder_interval(self) -> Tuple[int, int]:
         """
@@ -404,7 +436,7 @@ class RemoveBeforeMatch(SingleMatch):
         return slice(self.rstop, None)
 
     def trimmed(self, read):
-        return read[self.rstop:]
+        return read[self.rstop :]
 
     def removed_sequence_length(self) -> int:
         return self.rstop
@@ -413,10 +445,6 @@ class RemoveBeforeMatch(SingleMatch):
 class RemoveAfterMatch(SingleMatch):
     """A match that removes sequence after the match"""
 
-    def __repr__(self):
-        return "RemoveAfterMatch(astart={}, astop={}, rstart={}, rstop={}, matches={}, errors={})".format(
-            self.astart, self.astop, self.rstart, self.rstop, self.matches, self.errors)
-
     def rest(self) -> str:
         """
         Return the part of the read before this match if this is a
@@ -424,7 +452,7 @@ class RemoveAfterMatch(SingleMatch):
         return the part after the match if this is not a 'front' adapter (3').
         This can be an empty string.
         """
-        return self.sequence[self.rstop:]
+        return self.sequence[self.rstop :]
 
     def remainder_interval(self) -> Tuple[int, int]:
         """
@@ -441,10 +469,10 @@ class RemoveAfterMatch(SingleMatch):
         return slice(None, self.rstart)
 
     def trimmed(self, read):
-        return read[:self.rstart]
+        return read[: self.rstart]
 
     def adjacent_base(self) -> str:
-        return self.sequence[self.rstart - 1:self.rstart]
+        return self.sequence[self.rstart - 1 : self.rstart]
 
     def removed_sequence_length(self) -> int:
         return len(self.sequence) - self.rstart
@@ -459,7 +487,7 @@ def _generate_adapter_name(_start=[1]) -> str:
 class Matchable(ABC):
     """Something that has a match_to() method."""
 
-    def __init__(self, name: str, *args, **kwargs):
+    def __init__(self, name: Optional[str], *args, **kwargs):
         self.name = name
 
     @abstractmethod
@@ -501,7 +529,7 @@ class SingleAdapter(Adapter, ABC):
 
     max_errors -- Maximum allowed errors (non-negative float). If the values is less than 1, this is
         interpreted as a rate directly and passed to the aligner. If it is 1 or greater, the value
-        is converted to a rate by dividing it by the length of the sequence.
+        is converted to a rate by dividing it by the number of non-N characters in the sequence.
 
         The error rate is the number of errors in the alignment divided by the length
         of the part of the alignment that matches the adapter.
@@ -536,17 +564,19 @@ class SingleAdapter(Adapter, ABC):
         self.sequence: str = sequence.upper().replace("U", "T")
         if not self.sequence:
             raise ValueError("Adapter sequence is empty")
-        if max_errors >= 1:
-            max_errors /= len(self.sequence)
+        if max_errors >= 1 and self.sequence.count("N") != len(self.sequence):
+            max_errors /= len(self.sequence) - self.sequence.count("N")
         self.max_error_rate: float = max_errors
         self.min_overlap: int = min(min_overlap, len(self.sequence))
-        iupac = frozenset('ABCDGHKMNRSTUVWXY')
+        iupac = frozenset("ABCDGHKMNRSTUVWXY")
         if adapter_wildcards and not set(self.sequence) <= iupac:
             for c in self.sequence:
                 if c not in iupac:
                     if c == "I":
-                        extra = "For inosine, consider using N instead and please comment " \
-                                "on <https://github.com/marcelm/cutadapt/issues/546>."
+                        extra = (
+                            "For inosine, consider using N instead and please comment "
+                            "on <https://github.com/marcelm/cutadapt/issues/546>."
+                        )
                     else:
                         extra = "Use only characters 'ABCDGHKMNRSTUVWXY'."
                     raise InvalidCharacter(
@@ -554,17 +584,19 @@ class SingleAdapter(Adapter, ABC):
                         f"not a valid IUPAC code. {extra}"
                     )
         # Optimization: Use non-wildcard matching if only ACGT is used
-        self.adapter_wildcards: bool = adapter_wildcards and not set(self.sequence) <= set("ACGT")
+        self.adapter_wildcards: bool = adapter_wildcards and not set(
+            self.sequence
+        ) <= set("ACGT")
         self.read_wildcards: bool = read_wildcards
         self.indels: bool = indels
         self.aligner = self._aligner()
 
-    def _make_aligner(self, flags: int) -> align.Aligner:
+    def _make_aligner(self, flags: int) -> Aligner:
         # TODO
         # Indels are suppressed by setting their cost very high, but a different algorithm
         # should be used instead.
         indel_cost = 1 if self.indels else 100000
-        return align.Aligner(
+        return Aligner(
             self.sequence,
             self.max_error_rate,
             flags=flags,
@@ -575,11 +607,13 @@ class SingleAdapter(Adapter, ABC):
         )
 
     def __repr__(self):
-        return '<{cls}(name={name!r}, sequence={sequence!r}, '\
-            'max_error_rate={max_error_rate}, min_overlap={min_overlap}, '\
-            'read_wildcards={read_wildcards}, '\
-            'adapter_wildcards={adapter_wildcards}, '\
-            'indels={indels})>'.format(cls=self.__class__.__name__, **vars(self))
+        return (
+            "<{cls}(name={name!r}, sequence={sequence!r}, "
+            "max_error_rate={max_error_rate}, min_overlap={min_overlap}, "
+            "read_wildcards={read_wildcards}, "
+            "adapter_wildcards={adapter_wildcards}, "
+            "indels={indels})>".format(cls=self.__class__.__name__, **vars(self))
+        )
 
     @property
     def effective_length(self) -> int:
@@ -623,8 +657,10 @@ class FrontAdapter(SingleAdapter):
     def descriptive_identifier(self) -> str:
         return "regular_five_prime"
 
-    def _aligner(self) -> align.Aligner:
-        return self._make_aligner(Where.ANYWHERE.value if self._force_anywhere else Where.FRONT.value)
+    def _aligner(self) -> Aligner:
+        return self._make_aligner(
+            Where.ANYWHERE.value if self._force_anywhere else Where.FRONT.value
+        )
 
     def match_to(self, sequence: str):
         """
@@ -634,7 +670,9 @@ class FrontAdapter(SingleAdapter):
         return None if no match was found given the matching criteria (minimum
         overlap length, maximum error rate).
         """
-        alignment: Optional[Tuple[int, int, int, int, int, int]] = self.aligner.locate(sequence)
+        alignment: Optional[Tuple[int, int, int, int, int, int]] = self.aligner.locate(
+            sequence
+        )
         if self._debug:
             print(self.aligner.dpmatrix)
         if alignment is None:
@@ -648,6 +686,57 @@ class FrontAdapter(SingleAdapter):
         return FrontAdapterStatistics(self)
 
 
+class RightmostFrontAdapter(FrontAdapter):
+    """A 5' adapter that prefers rightmost matches"""
+
+    description = "rightmost 5'"
+
+    # def __init__(self, *args, **kwargs):
+    #     self._force_anywhere = kwargs.pop("force_anywhere", False)
+    #     super().__init__(*args, **kwargs)
+
+    def descriptive_identifier(self) -> str:
+        return "rightmost_five_prime"
+
+    def _aligner(self) -> Aligner:
+        self.sequence = self.sequence[::-1]
+        aligner = self._make_aligner(
+            Where.ANYWHERE.value if self._force_anywhere else Where.FRONT.value
+        )
+        self.sequence = self.sequence[::-1]
+        return aligner
+
+    def match_to(self, sequence: str):
+        """
+        Attempt to match this adapter to the given read.
+
+        Return a Match instance if a match was found;
+        return None if no match was found given the matching criteria (minimum
+        overlap length, maximum error rate).
+        """
+        alignment: Optional[Tuple[int, int, int, int, int, int]] = self.aligner.locate(
+            sequence[::-1]
+        )
+        if self._debug:
+            print(self.aligner.dpmatrix)
+        if alignment is None:
+            return None
+
+        ref_start, ref_end, query_start, query_end, score, errors = alignment
+        alignment = (
+            len(self.sequence) - ref_end,
+            len(self.sequence) - ref_start,
+            len(sequence) - query_end,
+            len(sequence) - query_start,
+            score,
+            errors,
+        )
+        return RemoveBeforeMatch(*alignment, adapter=self, sequence=sequence)
+
+    def spec(self) -> str:
+        return f"{self.sequence}...;rightmost"
+
+
 class BackAdapter(SingleAdapter):
     """A 3' adapter"""
 
@@ -661,7 +750,9 @@ class BackAdapter(SingleAdapter):
         return "regular_three_prime"
 
     def _aligner(self):
-        return self._make_aligner(Where.ANYWHERE.value if self._force_anywhere else Where.BACK.value)
+        return self._make_aligner(
+            Where.ANYWHERE.value if self._force_anywhere else Where.BACK.value
+        )
 
     def match_to(self, sequence: str):
         """
@@ -671,7 +762,9 @@ class BackAdapter(SingleAdapter):
         return None if no match was found given the matching criteria (minimum
         overlap length, maximum error rate).
         """
-        alignment: Optional[Tuple[int, int, int, int, int, int]] = self.aligner.locate(sequence)
+        alignment: Optional[Tuple[int, int, int, int, int, int]] = self.aligner.locate(
+            sequence
+        )
         if self._debug:
             print(self.aligner.dpmatrix)  # pragma: no cover
         if alignment is None:
@@ -795,12 +888,12 @@ class PrefixAdapter(NonInternalFrontAdapter):
 
     def _aligner(self):
         if not self.indels:  # TODO or if error rate allows 0 errors anyway
-            return align.PrefixComparer(
+            return PrefixComparer(
                 self.sequence,
                 self.max_error_rate,
                 wildcard_ref=self.adapter_wildcards,
                 wildcard_query=self.read_wildcards,
-                min_overlap=self.min_overlap
+                min_overlap=self.min_overlap,
             )
         else:
             return self._make_aligner(Where.PREFIX.value)
@@ -824,12 +917,12 @@ class SuffixAdapter(NonInternalBackAdapter):
 
     def _aligner(self):
         if not self.indels:  # TODO or if error rate allows 0 errors anyway
-            return align.SuffixComparer(
+            return SuffixComparer(
                 self.sequence,
                 self.max_error_rate,
                 wildcard_ref=self.adapter_wildcards,
                 wildcard_query=self.read_wildcards,
-                min_overlap=self.min_overlap
+                min_overlap=self.min_overlap,
             )
         else:
             return self._make_aligner(Where.SUFFIX.value)
@@ -842,25 +935,32 @@ class LinkedMatch(Match):
     """
     Represent a match of a LinkedAdapter
     """
-    def __init__(self, front_match: RemoveBeforeMatch, back_match: RemoveAfterMatch, adapter: "LinkedAdapter"):
+
+    def __init__(
+        self,
+        front_match: RemoveBeforeMatch,
+        back_match: RemoveAfterMatch,
+        adapter: "LinkedAdapter",
+    ):
         assert front_match is not None or back_match is not None
         self.front_match: RemoveBeforeMatch = front_match
         self.back_match: RemoveAfterMatch = back_match
         self.adapter: LinkedAdapter = adapter
 
     def __repr__(self):
-        return '<LinkedMatch(front_match={!r}, back_match={}, adapter={})>'.format(
-            self.front_match, self.back_match, self.adapter)
+        return "<LinkedMatch(front_match={!r}, back_match={}, adapter={})>".format(
+            self.front_match, self.back_match, self.adapter
+        )
 
     @property
-    def matches(self):
+    def score(self):
         """Number of matching bases"""
-        m = 0
+        s = 0
         if self.front_match is not None:
-            m += self.front_match.matches
+            s += self.front_match.score
         if self.back_match is not None:
-            m += self.back_match.matches
-        return m
+            s += self.back_match.score
+        return s
 
     @property
     def errors(self):
@@ -879,7 +979,9 @@ class LinkedMatch(Match):
         return read
 
     def remainder_interval(self) -> Tuple[int, int]:
-        matches = [match for match in [self.front_match, self.back_match] if match is not None]
+        matches = [
+            match for match in [self.front_match, self.back_match] if match is not None
+        ]
         return remainder(matches)
 
     def retained_adapter_interval(self) -> Tuple[int, int]:
@@ -903,11 +1005,20 @@ class LinkedMatch(Match):
             if match is None:
                 continue
             record = match.get_info_records(read)[0]
-            record[7] = ("none" if self.adapter.name is None else self.adapter.name) + namesuffix
+            record[7] = (
+                "none" if self.adapter.name is None else self.adapter.name
+            ) + namesuffix
             records.append(record)
             read = match.trimmed(read)
         return records
 
+    def match_sequence(self):
+        return (
+            (self.front_match.match_sequence() if self.front_match else "")
+            + ","
+            + (self.back_match.match_sequence() if self.back_match else "")
+        )
+
 
 class LinkedAdapter(Adapter):
     """A 5' adapter combined with a 3' adapter"""
@@ -920,15 +1031,15 @@ class LinkedAdapter(Adapter):
         back_adapter: SingleAdapter,
         front_required: bool,
         back_required: bool,
-        name: str,
+        name: Optional[str],
     ):
         super().__init__(name)
         self.front_required = front_required
         self.back_required = back_required
 
         # The following attributes are needed for the report
-        self.where = Where.LINKED
-        self.name = _generate_adapter_name() if name is None else name
+        self.where = "linked"
+        self.name: str = _generate_adapter_name() if name is None else name
         self.front_adapter = front_adapter
         self.front_adapter.name = self.name
         self.back_adapter = back_adapter
@@ -955,7 +1066,9 @@ class LinkedAdapter(Adapter):
         return LinkedMatch(front_match, back_match, self)
 
     def create_statistics(self) -> LinkedAdapterStatistics:
-        return LinkedAdapterStatistics(self, front=self.front_adapter, back=self.back_adapter)
+        return LinkedAdapterStatistics(
+            self, front=self.front_adapter, back=self.back_adapter
+        )
 
     @property
     def sequence(self):
@@ -973,6 +1086,7 @@ class MultipleAdapters(Matchable):
     """
     Represent multiple adapters at once
     """
+
     def __init__(self, adapters: Sequence[Matchable]):
         super().__init__(name="multiple_adapters")
         self._adapters = adapters
@@ -999,9 +1113,13 @@ class MultipleAdapters(Matchable):
             if match is None:
                 continue
 
-            # the no. of matches determines which adapter fits best
-            if best_match is None or match.matches > best_match.matches or (
-                match.matches == best_match.matches and match.errors < best_match.errors
+            # the score determines which adapter fits best
+            if (
+                best_match is None
+                or match.score > best_match.score
+                or (
+                    match.score == best_match.score and match.errors < best_match.errors
+                )
             ):
                 best_match = match
         return best_match
@@ -1020,6 +1138,7 @@ class IndexedAdapters(Matchable, ABC):
 
     Use the is_acceptable() method to check individual adapters.
     """
+
     AdapterIndex = Dict[str, Tuple[SingleAdapter, int, int]]
 
     def __init__(self, adapters):
@@ -1032,7 +1151,9 @@ class IndexedAdapters(Matchable, ABC):
         self._adapters = adapters
         self._multiple_adapters = MultipleAdapters(adapters)
         self._lengths, self._index = self._make_index()
-        logger.debug("String lengths in the index: %s", sorted(self._lengths, reverse=True))
+        logger.debug(
+            "String lengths in the index: %s", sorted(self._lengths, reverse=True)
+        )
         if len(self._lengths) == 1:
             self._length = self._lengths[0]
             self.match_to = self._match_to_one_length
@@ -1080,14 +1201,14 @@ class IndexedAdapters(Matchable, ABC):
         return True
 
     def _make_index(self) -> Tuple[List[int], "AdapterIndex"]:
-        logger.info('Building index of %s adapters ...', len(self._adapters))
+        logger.info("Building index of %s adapters ...", len(self._adapters))
         index: Dict[str, Tuple[SingleAdapter, int, int]] = dict()
         lengths = set()
         has_warned = False
         for adapter in self._adapters:
             sequence = adapter.sequence
             k = int(adapter.max_error_rate * len(sequence))
-            environment = align.edit_environment if adapter.indels else align.hamming_environment
+            environment = edit_environment if adapter.indels else hamming_environment
             for s, errors, matches in environment(sequence, k):
                 if s in index:
                     other_adapter, other_errors, other_matches = index[s]
@@ -1098,14 +1219,19 @@ class IndexedAdapters(Matchable, ABC):
                             "Adapters %s %r and %s %r are very similar. At %s allowed errors, "
                             "the sequence %r cannot be assigned uniquely because the number of "
                             "matches is %s compared to both adapters.",
-                            other_adapter.name, other_adapter.sequence, adapter.name,
-                            adapter.sequence, k, s, matches
+                            other_adapter.name,
+                            other_adapter.sequence,
+                            adapter.name,
+                            adapter.sequence,
+                            k,
+                            s,
+                            matches,
                         )
                         has_warned = True
                 else:
                     index[s] = (adapter, errors, matches)
                 lengths.add(len(s))
-        logger.info('Built an index containing %s strings.', len(index))
+        logger.info("Built an index containing %s strings.", len(index))
 
         return sorted(lengths, reverse=True), index
 
@@ -1168,20 +1294,19 @@ class IndexedAdapters(Matchable, ABC):
 
 
 class IndexedPrefixAdapters(IndexedAdapters):
-
     @classmethod
     def _accept(cls, adapter):
         if not isinstance(adapter, PrefixAdapter):
             raise ValueError("Only 5' anchored adapters are allowed")
         return super()._accept(adapter)
 
-    def _make_match(self, adapter, length, matches, errors, sequence):
+    def _make_match(self, adapter, length, score, errors, sequence):
         return RemoveBeforeMatch(
             astart=0,
             astop=len(adapter.sequence),
             rstart=0,
             rstop=length,
-            matches=matches,
+            score=score,
             errors=errors,
             adapter=adapter,
             sequence=sequence,
@@ -1196,20 +1321,19 @@ class IndexedPrefixAdapters(IndexedAdapters):
 
 
 class IndexedSuffixAdapters(IndexedAdapters):
-
     @classmethod
     def _accept(cls, adapter):
         if not isinstance(adapter, SuffixAdapter):
             raise ValueError("Only anchored 3' adapters are allowed")
         return super()._accept(adapter)
 
-    def _make_match(self, adapter, length, matches, errors, sequence):
+    def _make_match(self, adapter, length, score, errors, sequence):
         return RemoveAfterMatch(
             astart=0,
             astop=len(adapter.sequence),
             rstart=len(sequence) - length,
             rstop=len(sequence),
-            matches=matches,
+            score=score,
             errors=errors,
             adapter=adapter,
             sequence=sequence,
@@ -1228,9 +1352,12 @@ def warn_duplicate_adapters(adapters):
     for adapter in adapters:
         key = (adapter.__class__, adapter.sequence)
         if key in d:
-            logger.warning("Adapter %r (%s) was specified multiple times! "
+            logger.warning(
+                "Adapter %r (%s) was specified multiple times! "
                 "Please make sure that this is what you want.",
-                adapter.sequence, adapter.description)
+                adapter.sequence,
+                adapter.description,
+            )
         d[key] = adapter.name
 
 
