@@ -9,11 +9,18 @@ from types import SimpleNamespace
 from typing import Sequence, List, Tuple, Optional, Set
 from abc import ABC, abstractmethod
 
-from dnaio import record_names_match, Sequence as DnaSequence
+from dnaio import record_names_match, Sequence as SequenceRecord
 
 from .qualtrim import quality_trim_index, nextseq_trim_index
-from .adapters import MultipleAdapters, SingleAdapter, IndexedPrefixAdapters, IndexedSuffixAdapters, \
-    Match, remainder, Adapter
+from .adapters import (
+    MultipleAdapters,
+    SingleAdapter,
+    IndexedPrefixAdapters,
+    IndexedSuffixAdapters,
+    Match,
+    remainder,
+    Adapter,
+)
 from .tokenizer import tokenize_braces, TokenizeError, Token, BraceToken
 from .utils import reverse_complemented_sequence
 
@@ -30,10 +37,11 @@ class ModificationInfo:
     Any information (except the read itself) that needs to be passed from one modifier
     to one later in the pipeline or from one modifier to the filters is recorded here.
     """
+
     __slots__ = ["matches", "original_read", "cut_prefix", "cut_suffix", "is_rc"]
 
     def __init__(self, read):
-        self.matches = []  # type: List[Match]
+        self.matches: List[Match] = []
         self.original_read = read
         self.cut_prefix = None
         self.cut_suffix = None
@@ -60,17 +68,22 @@ class PairedEndModifier(ABC):
     @abstractmethod
     def __call__(
         self, read1, read2, info1: ModificationInfo, info2: ModificationInfo
-    ) -> Tuple[DnaSequence, DnaSequence]:
+    ) -> Tuple[SequenceRecord, SequenceRecord]:
         pass
 
 
 class PairedEndModifierWrapper(PairedEndModifier):
     """
-    Wrapper for modifiers that work on both reads in a paired-end read
+    Wrap two SingleEndModifiers that work on both reads in a paired-end read
     """
+
     paired = True
 
-    def __init__(self, modifier1: Optional[SingleEndModifier], modifier2: Optional[SingleEndModifier]):
+    def __init__(
+        self,
+        modifier1: Optional[SingleEndModifier],
+        modifier2: Optional[SingleEndModifier],
+    ):
         """Set one of the modifiers to None to work on R1 or R2 only"""
         self._modifier1 = modifier1
         self._modifier2 = modifier2
@@ -78,8 +91,10 @@ class PairedEndModifierWrapper(PairedEndModifier):
             raise ValueError("Not both modifiers may be None")
 
     def __repr__(self):
-        return "PairedEndModifierWrapper(" \
+        return (
+            "PairedEndModifierWrapper("
             f"modifier1={self._modifier1!r}, modifier2={self._modifier2!r})"
+        )
 
     def __call__(self, read1, read2, info1: ModificationInfo, info2: ModificationInfo):
         if self._modifier1 is None:
@@ -119,15 +134,21 @@ class AdapterCutter(SingleEndModifier):
         self.with_adapters = 0
         self.adapter_statistics = {a: a.create_statistics() for a in adapters}
         if index:
-            self.adapters = MultipleAdapters(self._regroup_into_indexed_adapters(adapters))
+            self.adapters = MultipleAdapters(
+                self._regroup_into_indexed_adapters(adapters)
+            )
         else:
             self.adapters = MultipleAdapters(adapters)
         if action == "retain" and times > 1:
             raise ValueError("'retain' cannot be combined with times > 1")
+        if self.times == 1 and self.action == "trim":
+            self.match_and_trim = self._match_and_trim_once_action_trim  # type: ignore
 
     def __repr__(self):
-        return "AdapterCutter(" \
+        return (
+            "AdapterCutter("
             f"adapters={self.adapters!r}, times={self.times}, action='{self.action}')"
+        )
 
     def _regroup_into_indexed_adapters(self, adapters):
         prefix, suffix, single = self._split_adapters(adapters)
@@ -149,8 +170,10 @@ class AdapterCutter(SingleEndModifier):
 
     @staticmethod
     def _split_adapters(
-        adapters: Sequence[SingleAdapter]
-    ) -> Tuple[Sequence[SingleAdapter], Sequence[SingleAdapter], Sequence[SingleAdapter]]:
+        adapters: Sequence[SingleAdapter],
+    ) -> Tuple[
+        Sequence[SingleAdapter], Sequence[SingleAdapter], Sequence[SingleAdapter]
+    ]:
         """
         Split adapters into three different categories so that they can possibly be used
         with a MultiAdapter. Return a tuple (prefix, suffix, other), where
@@ -158,9 +181,9 @@ class AdapterCutter(SingleEndModifier):
         - suffix is a list of all anchored 3' adapters that MultiAdapter would accept
         - other is a list of all remaining adapters.
         """
-        prefix = []  # type: List[SingleAdapter]
-        suffix = []  # type: List[SingleAdapter]
-        other = []  # type: List[SingleAdapter]
+        prefix: List[SingleAdapter] = []
+        suffix: List[SingleAdapter] = []
+        other: List[SingleAdapter] = []
         for a in adapters:
             if IndexedPrefixAdapters.is_acceptable(a):
                 prefix.append(a)
@@ -180,9 +203,8 @@ class AdapterCutter(SingleEndModifier):
         start, stop = remainder(matches)
         result = read[:]
         result.sequence = (
-            'N' * start
-            + read.sequence[start:stop]
-            + 'N' * (len(read) - stop))
+            "N" * start + read.sequence[start:stop] + "N" * (len(read) - stop)
+        )
         return result
 
     @staticmethod
@@ -218,7 +240,7 @@ class AdapterCutter(SingleEndModifier):
         Return a pair (trimmed_read, matches), where matches is a list of Match instances.
         """
         matches = []
-        if self.action == 'lowercase':  # TODO this should not be needed
+        if self.action == "lowercase":  # TODO this should not be needed
             read.sequence = read.sequence.upper()
         trimmed_read = read
         for _ in range(self.times):
@@ -232,14 +254,14 @@ class AdapterCutter(SingleEndModifier):
         if not matches:
             return trimmed_read, []
 
-        if self.action == 'trim':
+        if self.action == "trim":
             # read is already trimmed, nothing to do
             pass
-        elif self.action == 'retain':
+        elif self.action == "retain":
             trimmed_read = self.trim_but_retain_adapter(read, matches)
-        elif self.action == 'mask':
+        elif self.action == "mask":
             trimmed_read = self.masked_read(read, matches)
-        elif self.action == 'lowercase':
+        elif self.action == "lowercase":
             trimmed_read = self.lowercased_read(read, matches)
             assert len(trimmed_read.sequence) == len(read)
         elif self.action is None:
@@ -247,6 +269,16 @@ class AdapterCutter(SingleEndModifier):
 
         return trimmed_read, matches
 
+    def _match_and_trim_once_action_trim(self, read):
+        """
+        Specalization of match_and_trim for the case that self.times == 1 and self.action == 'trim'
+        """
+        match = self.adapters.match_to(read.sequence)
+        if match is not None:
+            return match.trimmed(read), [match]
+        else:
+            return read, []
+
 
 class ReverseComplementer(SingleEndModifier):
     """Trim adapters from a read and its reverse complement"""
@@ -266,11 +298,13 @@ class ReverseComplementer(SingleEndModifier):
         reverse_read = reverse_complemented_sequence(read)
 
         forward_trimmed_read, forward_matches = self.adapter_cutter.match_and_trim(read)
-        reverse_trimmed_read, reverse_matches = self.adapter_cutter.match_and_trim(reverse_read)
+        reverse_trimmed_read, reverse_matches = self.adapter_cutter.match_and_trim(
+            reverse_read
+        )
 
-        forward_match_count = sum(m.matches for m in forward_matches)
-        reverse_match_count = sum(m.matches for m in reverse_matches)
-        use_reverse_complement = reverse_match_count > forward_match_count
+        forward_score = sum(m.score for m in forward_matches)
+        reverse_score = sum(m.score for m in reverse_matches)
+        use_reverse_complement = reverse_score > forward_score
 
         if use_reverse_complement:
             self.reverse_complemented += 1
@@ -302,7 +336,7 @@ class PairedAdapterCutter(PairedEndModifier):
     Trim adapters in pairs from R1 and R2.
     """
 
-    def __init__(self, adapters1, adapters2, action='trim'):
+    def __init__(self, adapters1, adapters2, action="trim"):
         """
         adapters1 -- list of Adapters to be removed from R1
         adapters2 -- list of Adapters to be removed from R2
@@ -316,7 +350,8 @@ class PairedAdapterCutter(PairedEndModifier):
         if len(adapters1) != len(adapters2):
             raise PairedAdapterCutterError(
                 "The number of adapters to trim from R1 and R2 must be the same. "
-                "Given: {} for R1, {} for R2".format(len(adapters1), len(adapters2)))
+                "Given: {} for R1, {} for R2".format(len(adapters1), len(adapters2))
+            )
         if not adapters1:
             raise PairedAdapterCutterError("No adapters given")
         self._adapter_pairs = list(zip(adapters1, adapters2))
@@ -333,8 +368,7 @@ class PairedAdapterCutter(PairedEndModifier):
         return f"PairedAdapterCutter(adapter_pairs={self._adapter_pairs!r})"
 
     def __call__(self, read1, read2, info1, info2):
-        """
-        """
+        """ """
         best_matches = self._find_best_match_pair(read1.sequence, read2.sequence)
         if best_matches is None:
             return read1, read2
@@ -343,21 +377,21 @@ class PairedAdapterCutter(PairedEndModifier):
         result = []
         for i, match, read in zip([0, 1], [match1, match2], [read1, read2]):
             trimmed_read = read
-            if self.action == 'lowercase':
+            if self.action == "lowercase":
                 trimmed_read.sequence = trimmed_read.sequence.upper()
 
             trimmed_read = match.trimmed(trimmed_read)
             self.adapter_statistics[i][match.adapter].add_match(match)
 
-            if self.action == 'trim':
+            if self.action == "trim":
                 # read is already trimmed, nothing to do
                 pass
-            elif self.action == 'mask':
+            elif self.action == "mask":
                 trimmed_read = AdapterCutter.masked_read(read, [match])
-            elif self.action == 'lowercase':
+            elif self.action == "lowercase":
                 trimmed_read = AdapterCutter.lowercased_read(read, [match])
                 assert len(trimmed_read.sequence) == len(read)
-            elif self.action == 'retain':
+            elif self.action == "retain":
                 trimmed_read = AdapterCutter.trim_but_retain_adapter(read, [match])
             elif self.action is None:  # --no-trim
                 trimmed_read = read[:]
@@ -366,9 +400,11 @@ class PairedAdapterCutter(PairedEndModifier):
         info2.matches.append(match2)
         return result
 
-    def _find_best_match_pair(self, sequence1: str, sequence2: str) -> Optional[Tuple[Match, Match]]:
+    def _find_best_match_pair(
+        self, sequence1: str, sequence2: str
+    ) -> Optional[Tuple[Match, Match]]:
         best = None
-        best_matches = None
+        best_score = None
         best_errors = None
         for adapter1, adapter2 in self._adapter_pairs:
             match1 = adapter1.match_to(sequence1)
@@ -377,15 +413,15 @@ class PairedAdapterCutter(PairedEndModifier):
             match2 = adapter2.match_to(sequence2)
             if match2 is None:
                 continue
-            total_matches = match1.matches + match2.matches
+            total_score = match1.score + match2.score
             total_errors = match1.errors + match2.errors
             if (
                 best is None
-                or total_matches > best_matches
-                or (total_matches == best_matches and total_errors < best_errors)
+                or total_score > best_score
+                or (total_score == best_score and total_errors < best_errors)
             ):
                 best = match1, match2
-                best_matches = total_matches
+                best_score = total_score
                 best_errors = total_errors
         return best
 
@@ -397,6 +433,7 @@ class UnconditionalCutter(SingleEndModifier):
     If the length is positive, the bases are removed from the beginning of the read.
     If the length is negative, the bases are removed from the end of the read.
     """
+
     def __init__(self, length: int):
         self.length = length
 
@@ -405,17 +442,18 @@ class UnconditionalCutter(SingleEndModifier):
 
     def __call__(self, read, info: ModificationInfo):
         if self.length > 0:
-            info.cut_prefix = read.sequence[:self.length]
-            return read[self.length:]
+            info.cut_prefix = read.sequence[: self.length]
+            return read[self.length :]
         elif self.length < 0:
-            info.cut_suffix = read.sequence[self.length:]
-            return read[:self.length]
+            info.cut_suffix = read.sequence[self.length :]
+            return read[: self.length]
 
 
 class LengthTagModifier(SingleEndModifier):
     """
     Replace "length=..." strings in read names.
     """
+
     def __init__(self, length_tag):
         self.regex = re.compile(r"\b" + length_tag + r"[0-9]*\b")
         self.length_tag = length_tag
@@ -426,7 +464,9 @@ class LengthTagModifier(SingleEndModifier):
     def __call__(self, read, info: ModificationInfo):
         read = read[:]
         if read.name.find(self.length_tag) >= 0:
-            read.name = self.regex.sub(self.length_tag + str(len(read.sequence)), read.name)
+            read.name = self.regex.sub(
+                self.length_tag + str(len(read.sequence)), read.name
+            )
         return read
 
 
@@ -434,6 +474,7 @@ class SuffixRemover(SingleEndModifier):
     """
     Remove a given suffix from read names.
     """
+
     def __init__(self, suffix):
         self.suffix = suffix
 
@@ -443,7 +484,7 @@ class SuffixRemover(SingleEndModifier):
     def __call__(self, read, info: ModificationInfo):
         read = read[:]
         if read.name.endswith(self.suffix):
-            read.name = read.name[:-len(self.suffix)]
+            read.name = read.name[: -len(self.suffix)]
         return read
 
 
@@ -451,6 +492,7 @@ class PrefixSuffixAdder(SingleEndModifier):
     """
     Add a suffix and a prefix to read names
     """
+
     def __init__(self, prefix, suffix):
         self.prefix = prefix
         self.suffix = suffix
@@ -460,9 +502,12 @@ class PrefixSuffixAdder(SingleEndModifier):
 
     def __call__(self, read, info):
         read = read[:]
-        adapter_name = info.matches[-1].adapter.name if info.matches else 'no_adapter'
-        read.name = self.prefix.replace('{name}', adapter_name) + read.name + \
-            self.suffix.replace('{name}', adapter_name)
+        adapter_name = info.matches[-1].adapter.name if info.matches else "no_adapter"
+        read.name = (
+            self.prefix.replace("{name}", adapter_name)
+            + read.name
+            + self.suffix.replace("{name}", adapter_name)
+        )
         return read
 
 
@@ -482,8 +527,11 @@ class Renamer(SingleEndModifier):
     - {cut_prefix} -- prefix removed by UnconditionalCutter (with positive length argument)
     - {cut_suffix} -- suffix removed by UnconditionalCutter (with negative length argument)
     - {adapter_name} -- name of the *last* adapter match or no_adapter if there was none
+    - {match_sequence} -- the sequence that matched the adapter (this includes possible errors)
+          or an empty string if there was no match
     - {rc} -- the string 'rc' if the read was reverse complemented (with --revcomp) or '' otherwise
     """
+
     variables = {
         "header",
         "id",
@@ -492,6 +540,7 @@ class Renamer(SingleEndModifier):
         "cut_suffix",
         "adapter_name",
         "rc",
+        "match_sequence",
     }
 
     def __init__(self, template: str):
@@ -525,7 +574,7 @@ class Renamer(SingleEndModifier):
         else:
             return (read_name, "")
 
-    def __call__(self, read: DnaSequence, info: ModificationInfo) -> DnaSequence:
+    def __call__(self, read: SequenceRecord, info: ModificationInfo) -> SequenceRecord:
         id_, comment = self.parse_name(read.name)
         read.name = self._template.format(
             header=read.name,
@@ -533,8 +582,11 @@ class Renamer(SingleEndModifier):
             comment=comment,
             cut_prefix=info.cut_prefix if info.cut_prefix else "",
             cut_suffix=info.cut_suffix if info.cut_suffix else "",
-            adapter_name=info.matches[-1].adapter.name if info.matches else "no_adapter",
+            adapter_name=info.matches[-1].adapter.name
+            if info.matches
+            else "no_adapter",
             rc="rc" if info.is_rc else "",
+            match_sequence=info.matches[-1].match_sequence() if info.matches else "",
         )
         return read
 
@@ -570,8 +622,12 @@ class PairedEndRenamer(PairedEndModifier):
         return allowed
 
     def __call__(
-        self, read1: DnaSequence, read2: DnaSequence, info1: ModificationInfo, info2: ModificationInfo
-    ) -> Tuple[DnaSequence, DnaSequence]:
+        self,
+        read1: SequenceRecord,
+        read2: SequenceRecord,
+        info1: ModificationInfo,
+        info2: ModificationInfo,
+    ) -> Tuple[SequenceRecord, SequenceRecord]:
 
         id1, comment1 = Renamer.parse_name(read1.name)
         id2, comment2 = Renamer.parse_name(read2.name)
@@ -611,15 +667,23 @@ class PairedEndRenamer(PairedEndModifier):
     ) -> Tuple[str, str]:
         d = []
         for id_, comment, header, info in (
-            (id1, comment1, header1, info1), (id2, comment2, header2, info2)
+            (id1, comment1, header1, info1),
+            (id2, comment2, header2, info2),
         ):
+            if info.matches:
+                adapter_name = info.matches[-1].adapter.name
+                match_sequence = info.matches[-1].match_sequence()
+            else:
+                adapter_name = "no_adapter"
+                match_sequence = ""
             d.append(
                 dict(
                     comment=comment,
                     header=header,
                     cut_prefix=info.cut_prefix if info.cut_prefix else "",
                     cut_suffix=info.cut_suffix if info.cut_suffix else "",
-                    adapter_name=info.matches[-1].adapter.name if info.matches else "no_adapter",
+                    adapter_name=adapter_name,
+                    match_sequence=match_sequence,
                 )
             )
         name1 = self._template.format(
@@ -643,10 +707,11 @@ class ZeroCapper(SingleEndModifier):
     """
     Change negative quality values of a read to zero
     """
+
     def __init__(self, quality_base=33):
         self.quality_base = quality_base
         qb = quality_base
-        self.zero_cap_trans = str.maketrans(''.join(map(chr, range(qb))), chr(qb) * qb)
+        self.zero_cap_trans = str.maketrans("".join(map(chr, range(qb))), chr(qb) * qb)
 
     def __repr__(self):
         return f"ZeroCapper(quality_base={self.quality_base})"
@@ -658,7 +723,7 @@ class ZeroCapper(SingleEndModifier):
 
 
 class NextseqQualityTrimmer(SingleEndModifier):
-    def __init__(self, cutoff, base):
+    def __init__(self, cutoff: int, base: int = 33):
         self.cutoff = cutoff
         self.base = base
         self.trimmed_bases = 0
@@ -673,18 +738,22 @@ class NextseqQualityTrimmer(SingleEndModifier):
 
 
 class QualityTrimmer(SingleEndModifier):
-    def __init__(self, cutoff_front, cutoff_back, base):
+    def __init__(self, cutoff_front: int, cutoff_back: int, base: int = 33):
         self.cutoff_front = cutoff_front
         self.cutoff_back = cutoff_back
         self.base = base
         self.trimmed_bases = 0
 
     def __repr__(self):
-        return f"QualityTrimmer(cutoff_front={self.cutoff_front}, " \
-               f"cutoff_back={self.cutoff_back}, base={self.base})"
+        return (
+            f"QualityTrimmer(cutoff_front={self.cutoff_front}, "
+            f"cutoff_back={self.cutoff_back}, base={self.base})"
+        )
 
     def __call__(self, read, info: ModificationInfo):
-        start, stop = quality_trim_index(read.qualities, self.cutoff_front, self.cutoff_back, self.base)
+        start, stop = quality_trim_index(
+            read.qualities, self.cutoff_front, self.cutoff_back, self.base
+        )
         self.trimmed_bases += len(read) - (stop - start)
         return read[start:stop]
 
@@ -695,6 +764,7 @@ class Shortener(SingleEndModifier):
     If the length is positive, the bases are removed from the end of the read.
     If the length is negative, the bases are removed from the beginning of the read.
     """
+
     def __init__(self, length):
         self.length = length
 
@@ -703,16 +773,17 @@ class Shortener(SingleEndModifier):
 
     def __call__(self, read, info: ModificationInfo):
         if self.length >= 0:
-            return read[:self.length]
+            return read[: self.length]
         else:
-            return read[self.length:]
+            return read[self.length :]
 
 
 class NEndTrimmer(SingleEndModifier):
     """Trims Ns from the 3' and 5' end of reads"""
+
     def __init__(self):
-        self.start_trim = re.compile(r'^N+')
-        self.end_trim = re.compile(r'N+$')
+        self.start_trim = re.compile(r"^N+")
+        self.end_trim = re.compile(r"N+$")
 
     def __repr__(self):
         return "NEndTrimmer()"
