@@ -1,9 +1,8 @@
 import io
 import os
 import sys
-import copy
 import logging
-from typing import List, Optional, BinaryIO, TextIO, Any, Tuple, Dict
+from typing import List, Optional, BinaryIO, TextIO, Any, Tuple, Dict, Sequence, Union
 from abc import ABC, abstractmethod
 from multiprocessing import Process, Pipe, Queue
 from pathlib import Path
@@ -13,26 +12,57 @@ import traceback
 
 import dnaio
 
-from .utils import Progress, FileOpener
-from .modifiers import SingleEndModifier, PairedEndModifier, PairedEndModifierWrapper, ModificationInfo
+from .utils import Progress, FileOpener, open_raise_limit
+from .modifiers import (
+    SingleEndModifier,
+    PairedEndModifier,
+    PairedEndModifierWrapper,
+    ModificationInfo,
+)
 from .report import Statistics
-from .filters import (TooShort, TooLong, TooManyN, TooManyExpectedErrors, CasavaFiltered,
-    DiscardTrimmed, DiscardUntrimmed, Predicate)
-from .steps import SingleEndSink, PairedEndSink, SingleEndFilter, PairedEndFilter, Demultiplexer, \
-    PairedDemultiplexer, CombinatorialDemultiplexer, RestFileWriter, WildcardFileWriter, \
-    InfoFileWriter, SingleEndStep, PairedSingleEndStep
+from .filters import (
+    TooShort,
+    TooLong,
+    TooManyN,
+    TooManyExpectedErrors,
+    CasavaFiltered,
+    DiscardTrimmed,
+    DiscardUntrimmed,
+    Predicate,
+)
+from .steps import (
+    SingleEndSink,
+    PairedEndSink,
+    SingleEndFilter,
+    PairedEndFilter,
+    Demultiplexer,
+    PairedDemultiplexer,
+    CombinatorialDemultiplexer,
+    RestFileWriter,
+    WildcardFileWriter,
+    InfoFileWriter,
+    SingleEndStep,
+    PairedSingleEndStep,
+)
 
 logger = logging.getLogger()
 
 
 class InputFiles:
-    def __init__(self, file1: BinaryIO, file2: Optional[BinaryIO] = None, interleaved: bool = False):
+    def __init__(
+        self,
+        file1: BinaryIO,
+        file2: Optional[BinaryIO] = None,
+        interleaved: bool = False,
+    ):
         self.file1 = file1
         self.file2 = file2
         self.interleaved = interleaved
 
     def open(self):
-        return dnaio.open(self.file1, file2=self.file2, interleaved=self.interleaved, mode="r")
+        return dnaio.open(
+            self.file1, file2=self.file2, interleaved=self.interleaved, mode="r"
+        )
 
     def close(self) -> None:
         self.file1.close()
@@ -41,7 +71,9 @@ class InputFiles:
 
 
 class InputPaths:
-    def __init__(self, path1: str, path2: Optional[str] = None, interleaved: bool = False):
+    def __init__(
+        self, path1: str, path2: Optional[str] = None, interleaved: bool = False
+    ):
         self.path1 = path1
         self.path2 = path2
         self.interleaved = interleaved
@@ -57,6 +89,7 @@ class OutputFiles:
     and demultiplex_out2, which are dictionaries that map an adapter name
     to a file-like object.
     """
+
     def __init__(
         self,
         out: Optional[BinaryIO] = None,
@@ -110,8 +143,10 @@ class OutputFiles:
             if f is not None:
                 yield f
         for outs in (
-            self.demultiplex_out, self.demultiplex_out2,
-            self.combinatorial_out, self.combinatorial_out2,
+            self.demultiplex_out,
+            self.demultiplex_out2,
+            self.combinatorial_out,
+            self.combinatorial_out2,
         ):
             if outs is not None:
                 for f in outs.values():
@@ -124,12 +159,26 @@ class OutputFiles:
         """
         result = OutputFiles(force_fasta=self.force_fasta)
         for attr in (
-            "out", "out2", "untrimmed", "untrimmed2", "too_short", "too_short2", "too_long",
-            "too_long2", "info", "rest", "wildcard"
+            "out",
+            "out2",
+            "untrimmed",
+            "untrimmed2",
+            "too_short",
+            "too_short2",
+            "too_long",
+            "too_long2",
+            "info",
+            "rest",
+            "wildcard",
         ):
             if getattr(self, attr) is not None:
                 setattr(result, attr, io.BytesIO())
-        for attr in "demultiplex_out", "demultiplex_out2", "combinatorial_out", "combinatorial_out2":
+        for attr in (
+            "demultiplex_out",
+            "demultiplex_out2",
+            "combinatorial_out",
+            "combinatorial_out2",
+        ):
             if getattr(self, attr) is not None:
                 setattr(result, attr, dict())
                 for k, v in getattr(self, attr).items():
@@ -148,16 +197,17 @@ class Pipeline(ABC):
     """
     Processing pipeline that loops over reads and applies modifiers and filters
     """
+
     n_adapters = 0
     paired = False
 
-    def __init__(self, file_opener: FileOpener):
-        self._reader = None  # type: Any
-        self._steps = []  # type: List[Any]
-        self._infiles = None  # type: Optional[InputFiles]
-        self._outfiles = None  # type: Optional[OutputFiles]
+    def __init__(self):
+        self._reader: Any = None
+        self._steps: List[Any] = []
+        self._infiles: Optional[InputFiles] = None
+        self._outfiles: Optional[OutputFiles] = None
         self._demultiplexer = None
-        self._textiowrappers = []  # type: List[TextIO]
+        self._textiowrappers: List[TextIO] = []
 
         # Filter settings
         self._minimum_length = None
@@ -167,7 +217,6 @@ class Pipeline(ABC):
         self.discard_casava = False
         self.discard_trimmed = False
         self.discard_untrimmed = False
-        self.file_opener = file_opener
 
     def connect_io(self, infiles: InputFiles, outfiles: OutputFiles) -> None:
         self._infiles = infiles
@@ -196,12 +245,14 @@ class Pipeline(ABC):
             if outfile:
                 textiowrapper = io.TextIOWrapper(outfile)
                 self._textiowrappers.append(textiowrapper)
-                self._steps.append(self._wrap_single_end_step(step_class(textiowrapper)))
+                self._steps.append(
+                    self._wrap_single_end_step(step_class(textiowrapper))
+                )
 
         # minimum length and maximum length
         for lengths, file1, file2, predicate_class in (
             (self._minimum_length, outfiles.too_short, outfiles.too_short2, TooShort),
-            (self._maximum_length, outfiles.too_long, outfiles.too_long2, TooLong)
+            (self._maximum_length, outfiles.too_long, outfiles.too_long2, TooLong),
         ):
             if lengths is None:
                 continue
@@ -220,7 +271,8 @@ class Pipeline(ABC):
         if self.max_expected_errors is not None:
             if not self._reader.delivers_qualities:
                 logger.warning(
-                    "Ignoring option --max-ee because input does not contain quality values")
+                    "Ignoring option --max-ee because input does not contain quality values"
+                )
             else:
                 f1 = f2 = TooManyExpectedErrors(self.max_expected_errors)
                 self._steps.append(self._make_filter(None, f1, f2))
@@ -229,11 +281,21 @@ class Pipeline(ABC):
             f1 = f2 = CasavaFiltered()
             self._steps.append(self._make_filter(None, f1, f2))
 
-        if int(self.discard_trimmed) + int(self.discard_untrimmed) + int(outfiles.untrimmed is not None) > 1:
-            raise ValueError('discard_trimmed, discard_untrimmed and outfiles.untrimmed must not '
-                'be set simultaneously')
+        if (
+            int(self.discard_trimmed)
+            + int(self.discard_untrimmed)
+            + int(outfiles.untrimmed is not None)
+            > 1
+        ):
+            raise ValueError(
+                "discard_trimmed, discard_untrimmed and outfiles.untrimmed must not "
+                "be set simultaneously"
+            )
 
-        if outfiles.demultiplex_out is not None or outfiles.combinatorial_out is not None:
+        if (
+            outfiles.demultiplex_out is not None
+            or outfiles.combinatorial_out is not None
+        ):
             self._demultiplexer = self._create_demultiplexer(outfiles)
             self._steps.append(self._demultiplexer)
         else:
@@ -245,14 +307,15 @@ class Pipeline(ABC):
             # are mutually exclusive in order to avoid brain damage.
             if self.discard_trimmed:
                 self._steps.append(
-                    self._make_filter(None, DiscardTrimmed(), DiscardTrimmed()))
+                    self._make_filter(None, DiscardTrimmed(), DiscardTrimmed())
+                )
             elif self.discard_untrimmed:
-                self._steps.append(
-                    self._make_untrimmed_filter(None))
+                self._steps.append(self._make_untrimmed_filter(None))
             elif outfiles.untrimmed:
-                untrimmed_writer = self._open_writer(outfiles.untrimmed, outfiles.untrimmed2)
-                self._steps.append(
-                    self._make_untrimmed_filter(untrimmed_writer))
+                untrimmed_writer = self._open_writer(
+                    outfiles.untrimmed, outfiles.untrimmed2
+                )
+                self._steps.append(self._make_untrimmed_filter(untrimmed_writer))
 
             self._steps.append(self._final_filter(outfiles))
         for i, step in enumerate(self._steps, 1):
@@ -288,11 +351,15 @@ class Pipeline(ABC):
         return self._reader.delivers_qualities
 
     @abstractmethod
-    def process_reads(self, progress: Progress = None) -> Tuple[int, int, Optional[int]]:
+    def process_reads(
+        self, progress: Progress = None
+    ) -> Tuple[int, int, Optional[int]]:
         pass
 
     @abstractmethod
-    def _make_filter(self, writer, predicate1: Predicate, predicate2: Predicate):
+    def _make_filter(
+        self, writer, predicate1: Optional[Predicate], predicate2: Optional[Predicate]
+    ):
         pass
 
     @abstractmethod
@@ -316,16 +383,14 @@ class SingleEndPipeline(Pipeline):
     """
     Processing pipeline for single-end reads
     """
-    def __init__(self, file_opener: FileOpener):
-        super().__init__(file_opener)
-        self._modifiers = []  # type: List[SingleEndModifier]
 
-    def add(self, modifier: SingleEndModifier):
-        if modifier is None:
-            raise ValueError("Modifier must not be None")
-        self._modifiers.append(modifier)
+    def __init__(self, modifiers: List[SingleEndModifier]):
+        super().__init__()
+        self._modifiers: List[SingleEndModifier] = modifiers
 
-    def process_reads(self, progress: Progress = None) -> Tuple[int, int, Optional[int]]:
+    def process_reads(
+        self, progress: Progress = None
+    ) -> Tuple[int, int, Optional[int]]:
         """Run the pipeline. Return statistics"""
         n = 0  # no. of processed reads
         total_bp = 0
@@ -350,11 +415,19 @@ class SingleEndPipeline(Pipeline):
     ):
         assert file2 is None
         assert not isinstance(file, (str, bytes, Path))
-        return self.file_opener.dnaio_open_raise_limit(
-            file, mode="w", qualities=self.uses_qualities, fileformat="fasta" if force_fasta else None)
+        return open_raise_limit(
+            dnaio.open,
+            file,
+            mode="w",
+            qualities=self.uses_qualities,
+            fileformat="fasta" if force_fasta else None,
+        )
 
-    def _make_filter(self, writer, predicate1: Predicate, predicate2: Predicate):
+    def _make_filter(
+        self, writer, predicate1: Optional[Predicate], predicate2: Optional[Predicate]
+    ):
         _ = predicate2
+        assert predicate1 is not None
         return SingleEndFilter(writer, predicate1)
 
     def _make_untrimmed_filter(self, writer):
@@ -365,10 +438,12 @@ class SingleEndPipeline(Pipeline):
         writer = self._open_writer(outfiles.out, force_fasta=outfiles.force_fasta)
         return SingleEndSink(writer)
 
-    def _create_demultiplexer(self, outfiles: OutputFiles):
-        writers = dict()  # type: Dict[Optional[str], Any]
+    def _create_demultiplexer(self, outfiles: OutputFiles) -> Demultiplexer:
+        writers: Dict[Optional[str], Any] = dict()
         if outfiles.untrimmed is not None:
-            writers[None] = self._open_writer(outfiles.untrimmed, force_fasta=outfiles.force_fasta)
+            writers[None] = self._open_writer(
+                outfiles.untrimmed, force_fasta=outfiles.force_fasta
+            )
         assert outfiles.demultiplex_out is not None
         for name, file in outfiles.demultiplex_out.items():
             writers[name] = self._open_writer(file, force_fasta=outfiles.force_fasta)
@@ -400,37 +475,55 @@ class PairedEndPipeline(Pipeline):
     """
     Processing pipeline for paired-end reads.
     """
+
     paired = True
 
-    def __init__(self, file_opener: FileOpener, pair_filter_mode: str):
-        super().__init__(file_opener)
-        self._modifiers = []  # type: List[PairedEndModifier]
+    def __init__(
+        self,
+        modifiers: List[
+            Union[
+                PairedEndModifier,
+                Tuple[Optional[SingleEndModifier], Optional[SingleEndModifier]],
+            ]
+        ],
+        pair_filter_mode: str,
+    ):
+        super().__init__()
+        self._modifiers: List[PairedEndModifier] = []
         self._pair_filter_mode = pair_filter_mode
         self._reader = None
         # Whether to ignore pair_filter mode for discard-untrimmed filter
         self.override_untrimmed_pair_filter = False
+        self._add_modifiers(modifiers)
 
-    def add(self, modifier1: Optional[SingleEndModifier], modifier2: Optional[SingleEndModifier]) -> None:
+    def _add_modifiers(self, modifiers):
+        for modifier in modifiers:
+            if isinstance(modifier, tuple):
+                self._add_two_single_modifiers(*modifier)
+            else:
+                self._add_modifier(modifier)
+
+    def _add_two_single_modifiers(
+        self,
+        modifier1: Optional[SingleEndModifier],
+        modifier2: Optional[SingleEndModifier],
+    ) -> None:
         """
-        Add a modifier for R1 and R2. One of them can be None, in which case the modifier
-        will only be added for the respective read.
+        Add two single-end modifiers that modify R1 and R2, respectively.
+        One of them can be None, in which case the modifier
+        is only applied to the respective other read.
         """
         if modifier1 is None and modifier2 is None:
             raise ValueError("Not both modifiers can be None")
         self._modifiers.append(PairedEndModifierWrapper(modifier1, modifier2))
 
-    def add_both(self, modifier: SingleEndModifier) -> None:
-        """
-        Add one modifier for both R1 and R2
-        """
-        assert modifier is not None
-        self._modifiers.append(PairedEndModifierWrapper(modifier, copy.copy(modifier)))
-
-    def add_paired_modifier(self, modifier: PairedEndModifier) -> None:
+    def _add_modifier(self, modifier: PairedEndModifier) -> None:
         """Add a Modifier (without wrapping it in a PairedEndModifierWrapper)"""
         self._modifiers.append(modifier)
 
-    def process_reads(self, progress: Progress = None) -> Tuple[int, int, Optional[int]]:
+    def process_reads(
+        self, progress: Progress = None
+    ) -> Tuple[int, int, Optional[int]]:
         n = 0  # no. of processed reads
         total1_bp = 0
         total2_bp = 0
@@ -461,7 +554,8 @@ class PairedEndPipeline(Pipeline):
         # take care of threads and compression levels here.
         for f in (file, file2):
             assert not isinstance(f, (str, bytes, Path))
-        return self.file_opener.dnaio_open_raise_limit(
+        return open_raise_limit(
+            dnaio.open,
             file,
             file2=file2,
             mode="w",
@@ -470,10 +564,18 @@ class PairedEndPipeline(Pipeline):
             interleaved=file2 is None,
         )
 
-    def _make_filter(self, writer, predicate1: Predicate, predicate2: Predicate, pair_filter_mode=None):
+    def _make_filter(
+        self,
+        writer,
+        predicate1: Optional[Predicate],
+        predicate2: Optional[Predicate],
+        pair_filter_mode=None,
+    ):
         if pair_filter_mode is None:
             pair_filter_mode = self._pair_filter_mode
-        return PairedEndFilter(writer, predicate1, predicate2, pair_filter_mode=pair_filter_mode)
+        return PairedEndFilter(
+            writer, predicate1, predicate2, pair_filter_mode=pair_filter_mode
+        )
 
     def _make_untrimmed_filter(self, writer):
         """
@@ -484,11 +586,13 @@ class PairedEndPipeline(Pipeline):
             writer,
             DiscardUntrimmed(),
             DiscardUntrimmed(),
-            pair_filter_mode='both' if self.override_untrimmed_pair_filter else None,
+            pair_filter_mode="both" if self.override_untrimmed_pair_filter else None,
         )
 
     def _final_filter(self, outfiles):
-        writer = self._open_writer(outfiles.out, outfiles.out2, force_fasta=outfiles.force_fasta)
+        writer = self._open_writer(
+            outfiles.out, outfiles.out2, force_fasta=outfiles.force_fasta
+        )
         return PairedEndSink(writer)
 
     def _create_demultiplexer(self, outfiles):
@@ -533,7 +637,7 @@ class PairedEndPipeline(Pipeline):
 
 class ReaderProcess(Process):
     """
-    Read chunks of FASTA or FASTQ data (single-end or paired) and send to a worker.
+    Read chunks of FASTA or FASTQ data (single-end or paired) and send them to a worker.
 
     The reader repeatedly
 
@@ -544,7 +648,16 @@ class ReaderProcess(Process):
     and finally sends the stop token -1 ("poison pills") to all connections.
     """
 
-    def __init__(self, path: str, path2: Optional[str], opener: FileOpener, connections, queue, buffer_size, stdin_fd):
+    def __init__(
+        self,
+        path: str,
+        path2: Optional[str],
+        opener: FileOpener,
+        connections: Sequence[Connection],
+        queue: Queue,
+        buffer_size: int,
+        stdin_fd,
+    ):
         """
         queue -- a Queue of worker indices. A worker writes its own index into this
             queue to notify the reader that it is ready to receive more data.
@@ -564,20 +677,19 @@ class ReaderProcess(Process):
             sys.stdin.close()
             sys.stdin = os.fdopen(self.stdin_fd)
         try:
-            with self._opener.xopen(self.path, 'rb') as f:
+            with self._opener.xopen(self.path, "rb") as f:
                 if self.path2:
-                    with self._opener.xopen(self.path2, 'rb') as f2:
+                    with self._opener.xopen(self.path2, "rb") as f2:
                         for chunk_index, (chunk1, chunk2) in enumerate(
-                                dnaio.read_paired_chunks(f, f2, self.buffer_size)):
+                            dnaio.read_paired_chunks(f, f2, self.buffer_size)
+                        ):
                             self.send_to_worker(chunk_index, chunk1, chunk2)
                 else:
-                    for chunk_index, chunk in enumerate(dnaio.read_chunks(f, self.buffer_size)):
+                    for chunk_index, chunk in enumerate(
+                        dnaio.read_chunks(f, self.buffer_size)
+                    ):
                         self.send_to_worker(chunk_index, chunk)
-
-            # Send poison pills to all workers
-            for _ in range(len(self.connections)):
-                worker_index = self.queue.get()
-                self.connections[worker_index].send(-1)
+            self.shutdown()
         except Exception as e:
             # TODO better send this to a common "something went wrong" Queue
             for connection in self.connections:
@@ -592,6 +704,12 @@ class ReaderProcess(Process):
         if chunk2 is not None:
             connection.send_bytes(chunk2)
 
+    def shutdown(self):
+        # Send poison pills to all workers
+        for _ in range(len(self.connections)):
+            worker_index = self.queue.get()
+            self.connections[worker_index].send(-1)
+
 
 class WorkerProcess(Process):
     """
@@ -601,6 +719,7 @@ class WorkerProcess(Process):
     To notify the reader process that it wants data, it puts its own identifier into the
     need_work_queue before attempting to read data from the read_pipe.
     """
+
     def __init__(
         self,
         id_: int,
@@ -637,7 +756,7 @@ class WorkerProcess(Process):
                 elif chunk_index == -2:
                     # An exception has occurred in the reader
                     e, tb_str = self._read_pipe.recv()
-                    logger.error('%s', tb_str)
+                    logger.error("%s", tb_str)
                     raise e
 
                 infiles = self._make_input_files()
@@ -651,11 +770,14 @@ class WorkerProcess(Process):
                 self._pipeline.close()
 
             m = self._pipeline._modifiers
-            modifier_stats = Statistics().collect(0, 0, 0 if self._pipeline.paired else None, m, [])
+            modifier_stats = Statistics().collect(
+                0, 0, 0 if self._pipeline.paired else None, m, []
+            )
             stats += modifier_stats
             self._write_pipe.send(-1)
             self._write_pipe.send(stats)
         except Exception as e:
+            logger.error("%s", e)
             self._write_pipe.send(-2)
             self._write_pipe.send((e, traceback.format_exc()))
 
@@ -665,7 +787,7 @@ class WorkerProcess(Process):
 
         if self._two_input_files:
             data = self._read_pipe.recv_bytes()
-            input2 = io.BytesIO(data)  # type: Optional[BinaryIO]
+            input2: Optional[BinaryIO] = io.BytesIO(data)
         else:
             input2 = None
         return InputFiles(input, input2, interleaved=self._interleaved_input)
@@ -687,14 +809,14 @@ class OrderedChunkWriter:
     in any order. This class writes them to an output file in
     the correct order.
     """
+
     def __init__(self, outfile):
         self._chunks = dict()
         self._current_index = 0
         self._outfile = outfile
 
     def write(self, data, index):
-        """
-        """
+        """ """
         self._chunks[index] = data
         while self._current_index in self._chunks:
             self._outfile.write(self._chunks[self._current_index])
@@ -709,6 +831,7 @@ class PipelineRunner(ABC):
     """
     A read processing pipeline
     """
+
     def __init__(self, pipeline: Pipeline, progress: Progress):
         self._pipeline = pipeline
         self._progress = progress
@@ -763,7 +886,7 @@ class ParallelPipelineRunner(PipelineRunner):
     ):
         super().__init__(pipeline, progress)
         self._n_workers = n_workers
-        self._need_work_queue = Queue()  # type: Queue
+        self._need_work_queue: Queue = Queue()
         self._buffer_size = buffer_size
         self._outfiles = outfiles
         self._opener = opener
@@ -786,8 +909,15 @@ class ParallelPipelineRunner(PipelineRunner):
             # This happens during tests: pytest sets sys.stdin to an object
             # that does not have a file descriptor.
             fileno = -1
-        self._reader_process = ReaderProcess(path1, path2, self._opener, connw,
-            self._need_work_queue, self._buffer_size, fileno)
+        self._reader_process = ReaderProcess(
+            path1,
+            path2,
+            self._opener,
+            connw,
+            self._need_work_queue,
+            self._buffer_size,
+            fileno,
+        )
         self._reader_process.daemon = True
         self._reader_process.start()
 
@@ -798,10 +928,15 @@ class ParallelPipelineRunner(PipelineRunner):
             conn_r, conn_w = Pipe(duplex=False)
             connections.append(conn_r)
             worker = WorkerProcess(
-                index, self._pipeline,
+                index,
+                self._pipeline,
                 self._two_input_files,
-                self._interleaved_input, self._outfiles,
-                self._connections[index], conn_w, self._need_work_queue)
+                self._interleaved_input,
+                self._outfiles,
+                self._connections[index],
+                conn_w,
+                self._need_work_queue,
+            )
             worker.daemon = True
             worker.start()
             workers.append(worker)
@@ -815,37 +950,18 @@ class ParallelPipelineRunner(PipelineRunner):
         stats = Statistics()
         n = 0  # A running total of the number of processed reads (for progress indicator)
         while connections:
-            ready_connections = multiprocessing.connection.wait(connections)
-            for connection in ready_connections:  # type: Any
-                chunk_index = connection.recv()
+            ready_connections: List[Any] = multiprocessing.connection.wait(connections)
+            for connection in ready_connections:
+                chunk_index = self._try_receive(connection)
                 if chunk_index == -1:
                     # the worker is done
-                    cur_stats = connection.recv()
-                    if stats == -2:
-                        # An exception has occurred in the worker (see below,
-                        # this happens only when there is an exception sending
-                        # the statistics)
-                        e, tb_str = connection.recv()
-                        logger.error('%s', tb_str)
-                        raise e
+                    cur_stats = self._try_receive(connection)
                     stats += cur_stats
                     connections.remove(connection)
                     continue
-                elif chunk_index == -2:
-                    # An exception has occurred in the worker
-                    e, tb_str = connection.recv()
-
-                    # We should use the worker's actual traceback object
-                    # here, but traceback objects are not picklable.
-                    logger.error('%s', tb_str)
-                    raise e
 
                 # No. of reads processed in this chunk
-                chunk_n = connection.recv()
-                if chunk_n == -2:
-                    e, tb_str = connection.recv()
-                    logger.error('%s', tb_str)
-                    raise e
+                chunk_n = self._try_receive(connection)
                 n += chunk_n
                 self._progress.update(n)
                 for writer in writers:
@@ -859,6 +975,24 @@ class ParallelPipelineRunner(PipelineRunner):
         self._progress.stop(n)
         return stats
 
+    @staticmethod
+    def _try_receive(connection):
+        """
+        Try to receive data over self.connection and return it.
+        If an exception was received, raise it.
+        """
+        result = connection.recv()
+        if result == -2:
+            # An exception has occurred on the other end
+            e, tb_str = connection.recv()
+            # Under some conditions, the Make sure the error is printed
+            logger.error("%s", e)
+            # The other end does not send an actual traceback object because these are
+            # not picklable, but a string representation.
+            logger.debug("%s", tb_str)
+            raise e
+        return result
+
     def close(self) -> None:
         self._outfiles.close()
 
@@ -879,13 +1013,17 @@ class SerialPipelineRunner(PipelineRunner):
         self._pipeline.connect_io(infiles, outfiles)
 
     def run(self) -> Statistics:
-        (n, total1_bp, total2_bp) = self._pipeline.process_reads(progress=self._progress)
+        (n, total1_bp, total2_bp) = self._pipeline.process_reads(
+            progress=self._progress
+        )
         if self._progress:
             self._progress.stop(n)
         # TODO
         modifiers = getattr(self._pipeline, "_modifiers", None)
         assert modifiers is not None
-        return Statistics().collect(n, total1_bp, total2_bp, modifiers, self._pipeline._steps)
+        return Statistics().collect(
+            n, total1_bp, total2_bp, modifiers, self._pipeline._steps
+        )
 
     def close(self) -> None:
         self._pipeline.close()
